import cPickle as pickle
from os.path import join
import os, sys
import multiprocessing as mp
import time
import json
import csv
import pandas as pd
csv.field_size_limit(sys.maxsize)

def merge_dicts(*dict_args):
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

def csv_length():
    """
    Function that returns the number of non-redundant sequences that parsed ANARCI
    """
    df = pd.read_csv( join("Temp_CSV","temporary.csv") , sep=",", header=None)
    return len(df)

def read_csv(chunk,chunk_size):
    """Function that reads the temporary CSV file in chunks of chunk_size
    Inputs:
            @chunk: # of chunks
            @chunk_size: Size of chunks. Default is 250,000 sequences
    Return:
            @csv_dictonary: Python dictionary, where key is antibody sequence, and value is the numbered antibody residues
    """
    csv_dictonary = {}
    count = 0
    start = chunk*chunk_size

    with open(join("Temp_CSV","temporary.csv")) as csvDataFile:
            csvReader = csv.reader(csvDataFile)
            for row in csvReader:
                    if count >= start:
                            csv_dictonary[row[0]] = row[4]
                    if len(csv_dictonary) == chunk_size:
                            return csv_dictonary
                    count += 1	
    return csv_dictonary

def check_chain_correctness(pos_num, amino, ReferenceChain):
    """Function that checks if a given amino/position is present in the reference matrix. (0 means it is absent)
    Inputs:
            @pos_num: Position in antibody chain
            @amino: amino acid (or residue)
            @ReferenceChain: Reference matrix
    Return:
            True or False if a amino/position is present in the reference matrix

    """
    if pos_num in ReferenceChain:
        if ReferenceChain[pos_num][amino] != 0:
                return True
        else:
                return False
    else:
        return False


def checking_sequence_correctness(numbers,reference):
	"""Function runs every residue/position in a numbered antibody sequence against the reference matrix. If a residue/position not found in the refence matrix it is flagged.
	Inputs:
		@numbers:   Numbered antibody residue/positions separated by an antibody region (e.g. fwh1)
		@reference: Reference Matrix
	Return:
		@current_error: Number of ABOSS flagged residue/positions in a sequence
		@incorrect_positions: ABOSS flagged residue/positions in a sequence
	"""
	incorrect_positions = []
	current_error = 0
	for frame in numbers:
            for pos_num in numbers[frame]:
                if numbers[frame][pos_num] == "-":
                        continue
                else:
                    if int(pos_num[:3]) < 27:
                        if not check_chain_correctness(pos_num, numbers[frame][pos_num], reference["f1"]):
                            current_error +=1
                            incorrect_positions.append((pos_num, numbers[frame][pos_num]))

                    elif int(pos_num[:3]) < 39:
                        if not check_chain_correctness(pos_num, numbers[frame][pos_num], reference["c1"]):
                            current_error +=1
                            incorrect_positions.append((pos_num, numbers[frame][pos_num]))

                    elif int(pos_num[:3]) < 56:
                        if not check_chain_correctness(pos_num, numbers[frame][pos_num], reference["f2"]):
                            current_error +=1
                            incorrect_positions.append((pos_num, numbers[frame][pos_num]))

                    elif int(pos_num[:3]) < 66:
                        if not check_chain_correctness(pos_num, numbers[frame][pos_num], reference["c2"]):
                            current_error +=1
                            incorrect_positions.append((pos_num, numbers[frame][pos_num]))

                    elif int(pos_num[:3]) < 105:
                        if not check_chain_correctness(pos_num, numbers[frame][pos_num], reference["f3"]):
                            current_error +=1
                            incorrect_positions.append((pos_num, numbers[frame][pos_num]))

                    elif int(pos_num[:3]) < 118:
                        if not check_chain_correctness(pos_num, numbers[frame][pos_num], reference["c3"]):
                            current_error +=1
                            incorrect_positions.append((pos_num, numbers[frame][pos_num]))

                    elif int(pos_num[:3]) >= 118:
                        if not check_chain_correctness(pos_num, numbers[frame][pos_num], reference["f4"]):
                            current_error +=1
                            incorrect_positions.append((pos_num, numbers[frame][pos_num]))
	if incorrect_positions == []:
		incorrect_positions.append(None)

	return current_error, incorrect_positions

def initiating_ref_matrix(passed_new_lines,core, reference_matrix, originalDataset):
        """Passing numbered antibody sequences to a function that ABOSS flags antibody sequences
        Inputs:
                @passed_new_lines: python dictinary where key is antibody sequence, and value is the numbered antibody residues
                @core: core number to track the progress
        Return:
                @ABOSSParsedSequences: Sequences with ABOSS flagged reisdue/positions
        """
        ABOSSParsedSequences = {}
        for nonRedundantSequence in passed_new_lines:
                aa = json.loads(passed_new_lines[nonRedundantSequence])
                error, incorrect_residues = checking_sequence_correctness(aa, reference_matrix)  # Here I pass sequence to a reference matrix
                if error not in ABOSSParsedSequences:
                        ABOSSParsedSequences[error]= {}
                if nonRedundantSequence in ABOSSParsedSequences[error]:
                        continue
                else:
                    if isinstance(originalDataset[nonRedundantSequence], int):
                        multiplication_factor = originalDataset[nonRedundantSequence]
                    else:
                        multiplication_factor = len(originalDataset[nonRedundantSequence])
                    ABOSSParsedSequences[error][nonRedundantSequence] = (multiplication_factor, incorrect_residues)
        print "Portion %d has finished" %(core+1)
        return ABOSSParsedSequences

class FlagResiduePositions(object):
	""" class that goes through all sequences in Ig-Seq data, and checks if residue/positions are above the residue error rate.
	Inputs:
		@cores: number of cores used in aboss analysis
		@startingData: python dictionary,key are antibody sequences, values are corresponding redundancies
		@antibodychain: H or L
	Outputs:
		@all_regions_aa: Counts of amino acids at antibody positions in Ig-seq data separated by the number of ABOSS flagged residue/positions
		@correct_sequences: Original antibody sequences separated by the number of ABOSS flagged residue/positions
	"""

        def __init__(self, cores, originalDataset, chain):
            self.cores = cores
            self.originalDataset= originalDataset
            self.chain = chain
            self.reference_matrix =  self.loadRefMatrix()

            self.all_regions_aa = {}
            self.correct_sequences = {}

        def loadRefMatrix(self):
            reference_matrix = {}
            
            chainfo = {"H":{"f":"fwh","c":"cdrh"},
                       "L":{"f":"fwl","c":"cdrl"}}

            for region in ["f1","f2","f3","f4","c1","c2","c3"]:
                reg = "{0}{1}".format(chainfo[self.chain][region[0]], region[1])
                reference_matrix[region] = pickle.load(open("Reference_matrix/reference_matrix_{0}.pkl".format(reg),"rb")).transpose()
            return reference_matrix

        def recordResiduePositions(self, filtered_sequences_portion, chunkNumberedSequences, error):

            for sequence in filtered_sequences_portion:
                    aa = json.loads(chunkNumberedSequences[sequence])

                    if isinstance(self.originalDataset[sequence], int):
                            multiplication_factor = self.originalDataset[sequence]
                    else:
                            multiplication_factor = len(self.originalDataset[sequence])
                    for framework in aa:
                        if framework not in self.all_regions_aa[error]:
                            self.all_regions_aa[error][framework] = {}
                        for pos in aa[framework]:
                            for ele in aa[framework][pos]:
                                if pos not in self.all_regions_aa[error][framework]:
                                        self.all_regions_aa[error][framework][pos] = {}
                                if ele not in self.all_regions_aa[error][framework][pos]:
                                    self.all_regions_aa[error][framework][pos][ele] = 0
                                self.all_regions_aa[error][framework][pos][ele] += multiplication_factor


        def flagsequences(self):
            correct_sequences = {}
            t1 = time.time()
            temp_csv_length = csv_length()
            chunk_size = 250000
            number_of_chunks = temp_csv_length/ chunk_size +1
            for chunk in range(number_of_chunks):

                chunkNumberedSequences = read_csv(chunk,chunk_size)
                print "The current analysed chunk is {0} of out {1}".format(int(chunk)+1,number_of_chunks)
                num_of_entries = len(chunkNumberedSequences)
                print "The total number of non-redundant sequences in the chunk: {0}".format(num_of_entries)

                # Multiprocessing
                pool = mp.Pool(processes=self.cores)
                items_sort = chunkNumberedSequences.items()
                results = [ pool.apply_async(initiating_ref_matrix, args = (dict(items_sort[i::self.cores]),i, self.reference_matrix, self.originalDataset )) for i in xrange(self.cores) ]
                del items_sort
                pool.close()
                
                # Collecting outputs 
                filtered_sequences_portion = [ r.get() for r in results ]
                del results			

                # Going through portions
                for portion in range(len(filtered_sequences_portion)):
                    for error in filtered_sequences_portion[portion]:

                        #  Recording errors
                        if error not in self.all_regions_aa:
                                self.all_regions_aa[error] = {}
                                self.correct_sequences[error] = {}

                        # Recording ANARCI parsed sequences with different numbers of errors
                        self.correct_sequences[error] = merge_dicts(self.correct_sequences[error], filtered_sequences_portion[portion][error])

                        # Recording residue positions 
                        self.recordResiduePositions( filtered_sequences_portion[portion][error], chunkNumberedSequences, error )

                print "Time took to analyse this file: ",time.time() - t1

                del filtered_sequences_portion
                del chunkNumberedSequences

            return self.all_regions_aa, self.correct_sequences

if __name__ == '__main__':
	pass
