import pickle
import matplotlib.pyplot as plt
import numpy as np
from os.path import join


def adding_one(x):
    return x+1

def findError():
    with open("Error/ig_seq_residue_error_rate.txt", "r") as d:
        lines = d.readlines()
        residue_error_rate = lines[1].split(" ")[-2]
        return residue_error_rate

def plotBoxplot(data,dict_repeats,anarci_parsed, residue_error_rate):

    fig, axis = plt.subplots(figsize=(10,8))
    plt.boxplot(data)
    plt.axhline(y=anarci_parsed * float(residue_error_rate), xmin=0, xmax=1, hold=None)
    plt.xticks(map(adding_one, range(len(dict_repeats))),range(len(dict_repeats)))
    axis.set_ylabel("Number of repeats", fontsize=16)
    axis.set_xlabel("Number of flagged residue/positions", fontsize=16)
    plt.figtext(.5,0.9,"Number of flagged residue/positions versus sequence redundancy",fontsize=16,ha='center')
    plt.savefig("Graphs/IncorrectVSredundancy", dpi= 300)
    plt.clf()

def calculateAnarciParsed(non_redundant_reads):

    anarci_parsed = 0
    for error in non_redundant_reads:
         for num in non_redundant_reads[error].values():
             anarci_parsed += num[0]
    return anarci_parsed

def PlotRedundancyPlot(non_redundant_reads, residue_error_rate=False):

    # Redundandacies vs Flags
    dict_repeats = {}
    for error in non_redundant_reads: 
            if error not in dict_repeats:
                    dict_repeats[ error ] = {}

            for read in non_redundant_reads[error]:

                    copy = non_redundant_reads[error][read][0]

                    if copy not in dict_repeats[error]:
                            dict_repeats[error][copy] = 0
                    dict_repeats[error][copy] += 1
    data  = []
    for u in dict_repeats:
            data.append(sum([[x]*dict_repeats[u][x] for x in dict_repeats[u]], []))
    
    # Loading residue error rate
    if not residue_error_rate:
        residue_error_rate = plotBoxplot()

    # Calculating ANARCI parsed
    anarci_parsed = calculateAnarciParsed(non_redundant_reads)

    # plotting
    plotBoxplot(data, dict_repeats, anarci_parsed, residue_error_rate)

    return;

def calculate_filtered_reads(dataset,passed_reads_polished):
	print "\tProportion of non-redundant sequences that were filtered by ABOSS: {0}".format(float(len(passed_reads_polished))/len(dataset))
	if isinstance(dataset.items()[0][1], int):
            print "\tProportion of redundant sequences that were filtered by ABOSS: {0}".format(float(sum([passed_reads_polished[x][0] for x in passed_reads_polished]))/sum(dataset.values()))
	else:
            redundant_dataset_length = sum([len(dataset[ss]) for ss in dataset])
            print "\tProportion of redundant sequences that were filtered by ABOSS: {0}".format(float(sum([passed_reads_polished[x][0] for x in passed_reads_polished]))/redundant_dataset_length)
	return;

if __name__ == '__main__':
	pass

